# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import functools
import re
import time
from collections import defaultdict
from typing import List, NamedTuple, Iterator, Optional

from toolz import concat, take

from smily.util import ignore_nones


class LogEvent(NamedTuple):
    message: str
    timestamp: int
    ingestion_time: int

    @classmethod
    def from_raw(cls, raw):
        return cls(
            message=raw["message"],
            timestamp=raw["timestamp"],
            ingestion_time=raw["ingestionTime"],
        )

    def __str__(self):
        return self.message


class LogFilterEvent(NamedTuple):
    message: str
    timestamp: int
    ingestion_time: int
    # additional fields compared to `LogEvent`
    stream_name: str
    event_id: str

    @classmethod
    def from_raw(cls, raw):
        return cls(
            message=raw["message"],
            timestamp=raw["timestamp"],
            ingestion_time=raw["ingestionTime"],
            stream_name=raw["logStreamName"],
            event_id=raw["eventId"],
        )

    def __str__(self):
        return self.message


class LogStream(NamedTuple):
    logs: object
    group: str
    stream_name: str

    @property
    def _kwargs(self):
        return {"logGroupName": self.group, "logStreamName": self.stream_name}

    def iter_log_events(
        self, start_at_top=None, limit=None, forever=False
    ) -> Iterator[dict]:
        """Stream over log-events.

        This method is used by all other methods to access logs.
        """

        get_log_events = functools.partial(
            ignore_nones(self.logs.get_log_events),
            startFromHead=start_at_top,
            limit=limit,
            **self._kwargs,
        )

        token = None

        while True:
            response = get_log_events(nextToken=token)
            yield map(LogEvent.from_raw, response["events"])

            if token == response["nextForwardToken"]:
                if forever:
                    yield None
                else:
                    break

            token = response["nextForwardToken"]

    def head(self, n):
        assert n <= 10_000
        stream = concat(self.iter_log_events(start_at_top=True, limit=n))
        yield from take(n, stream)

    def tail(self, n):
        assert n <= 10_000
        stream = concat(self.iter_log_events(start_at_top=False, limit=n))
        yield from take(n, stream)

    def cat(self):
        stream = concat(self.iter_log_events(start_at_top=True))
        yield from stream

    def follow(self, start_at_top=True, limit=None, interval=30):
        for page in self.iter_log_events(
            start_at_top=start_at_top, forever=True, limit=limit
        ):
            if page is None:
                time.sleep(interval)
                continue
            else:
                yield from page


class LogStreamFilter(NamedTuple):
    logs: object
    group: str
    stream_names: Optional[List[str]] = None
    prefix: Optional[str] = None

    def iter_log_events(self, pattern: str, limit=None):
        assert (
            self.stream_names is None
            or self.prefix is None
            and self.stream_names != self.prefix
        ), "Provide either `stream_names` or `prefix`, and not both."

        filter_log_events = functools.partial(
            ignore_nones(self.logs.filter_log_events),
            logGroupName=self.group,
            logStreamNamePrefix=self.prefix,
            logStreamNames=self.stream_names,
            filterPattern=pattern,
            limit=limit,
        )

        token = None

        while True:
            response = filter_log_events(nextToken=token)
            token = response.get("nextToken")
            yield map(LogFilterEvent.from_raw, response["events"])

            if not token:
                break

    def iter_filter(self, pattern: str):
        for page in self.iter_log_events(pattern):
            yield from page

    def filter(self, pattern: str) -> dict:
        events_by_job = defaultdict(list)

        for event in self.iter_filter(pattern):
            events_by_job[event.stream_name].append(event)
        return events_by_job

    def iter_match(self, pattern: str, regex: str):
        for page in self.iter_log_events(pattern):
            for event in page:
                match = re.match(regex, event.message)
                if match is not None:
                    yield match, event

    def match(self, pattern: str, regex: str) -> dict:
        events_by_job = defaultdict(list)

        for match, event in self.iter_match(pattern, regex):
            events_by_job[event.stream_name].append(match.group(0))

        return events_by_job

    def iter_find(self, pattern: str, regex: str):
        for page in self.iter_log_events(pattern):
            for event in page:
                results = re.findall(regex, event.message)
                if results:
                    yield results, event

    def find(self, pattern: str, regex: str):
        events_by_job = defaultdict(list)

        for match, event in self.iter_find(pattern, regex):
            events_by_job[event.stream_name].append(match[0])

        return events_by_job


class TextFilter(NamedTuple):
    text: str

    def iter_log_events(self, pattern: str, limit=None):
        yield [
            LogFilterEvent(
                message=line,
                timestamp=idx,
                ingestion_time=idx,
                stream_name="",
                event_id=str(idx),
            )
            for idx, line in enumerate(self.text.splitlines())
        ]

    def iter_filter(self, pattern: str):
        for page in self.iter_log_events(pattern):
            yield from page

    def filter(self, pattern: str) -> dict:
        events_by_job = defaultdict(list)

        for event in self.iter_filter(pattern):
            events_by_job[event.stream_name].append(event)
        return events_by_job

    def iter_match(self, pattern: str, regex: str):
        for page in self.iter_log_events(pattern):
            for event in page:
                match = re.match(regex, event.message)
                if match is not None:
                    yield match, event

    def match(self, pattern: str, regex: str) -> dict:
        events_by_job = defaultdict(list)

        for match, event in self.iter_match(pattern, regex):
            events_by_job[event.stream_name].append(match.group(0))

        return events_by_job

    def iter_find(self, pattern: str, regex: str):
        for page in self.iter_log_events(pattern):
            for event in page:
                results = re.findall(regex, event.message)
                if results:
                    yield results, event

    def find(self, pattern: str, regex: str):
        events_by_job = defaultdict(list)

        for match, event in self.iter_find(pattern, regex):
            events_by_job[event.stream_name].append(match[0])

        return events_by_job



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import json
from pathlib import Path, PurePath
from urllib.parse import urlparse

import click
from toolz import valmap

from smily import env
from ._base import main, Resource


def get_data_channels(description):
    return {
        channel["ChannelName"]: channel["DataSource"]["S3DataSource"]["S3Uri"]
        for channel in description["InputDataConfig"]
    }


def download_files(bucket, prefix, destination):
    for key_object in bucket.objects.filter(Prefix=prefix.lstrip("/")):
        path = PurePath(key_object.key)

        bucket.download_file(str(path), str(destination / path.name))


@main.command()
@click.option("--arn", type=Resource)
@click.argument("destination", type=click.Path())
def scaffold(arn, destination):
    """Generate a scaffold for a training job."""
    if arn is not None:
        hyperparameters = arn._description["HyperParameters"]
        channels = get_data_channels(arn._description)
    else:
        hyperparameters = {}
        channels = {"train": None, "test": None}

    destination = Path(destination)
    make_config(destination, hyperparameters, channels)
    download_data(destination, channels)


def make_config(destination, hyperparameters, channels):
    config = destination / "input" / "config"
    config.mkdir(parents=True, exist_ok=True)

    # hyperparameters
    hyperparameters = valmap(str, hyperparameters)
    with open(config / "hyperparameters.json", "w") as hp_file:
        json.dump(hyperparameters, hp_file)

    # inputdataconfig
    input_data_config = {
        channel: {"ContentType": "auto"} for channel in channels
    }
    with open(config / "inputdataconfig.json", "w") as idc_file:
        json.dump(input_data_config, idc_file)


def download_data(destination, channels):
    data = destination / "input" / "data"
    data.mkdir(parents=True, exist_ok=True)

    for channel_name, s3path in channels.items():
        channel_path = data / channel_name
        channel_path.mkdir(exist_ok=True)

        if s3path is not None:
            download_files(s3path, channel_path)


def download_files(s3path, channel_path):
    s3 = env.boto_session.resource("s3")

    url = urlparse(s3path)
    bucket = s3.Bucket(url.netloc)
    path = url.path.lstrip("/")

    for path in bucket.objects.filter(Prefix=path):
        name = PurePath(path.key).name

        bucket.download_file(path.key, str(channel_path / name))



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import configparser
import re
import os
from pathlib import Path

from toolz import valmap

ACCOUNT_ID_RE = re.compile(r"\b(\d{12})\b")

config_file = Path(os.environ.get("AWS_CONFIG_FILE", "~/.aws/config")).expanduser()


def _get_profiles(config):
    for key, settings in config.items():
        if key == "default":
            yield key, dict(settings)

        elif key.startswith("profile "):
            name = key.split(" ", 1)[1]
            yield name, dict(settings)


def _find_account_id(settings):
    for value in settings.values():
        for match in ACCOUNT_ID_RE.finditer(value):
            return match.group(0)


def get_account_for_profiles(profiles):
    accounts = {}

    for profile, settings in profiles.items():
        account = _find_account_id(settings)
        if account is not None and account not in accounts:
            accounts[account] = profile

    return accounts


class Config:
    @classmethod
    def read(cls, path=config_file):
        path = Path(path)

        config = configparser.ConfigParser()

        if path.is_file():
            config.read(path)
        return cls(path, config)

    def __init__(self, path, config):
        self.path = path
        self.config = config
        self.profiles = dict(_get_profiles(config))
        self.accounts = get_account_for_profiles(self.profiles)



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import sys


def main():
    try:
        import click
    except ImportError:
        print(
            "`click` not installed. Try `pip installl smily[cli]` to ensure all "
            "dependencies are installed."
        )
        sys.exit(1)

    from .cli import main

    main()


if __name__ == "__main__":
    main()



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import boto3

from dataclasses import dataclass
from typing import Optional

from smily.lib import Resource, TrainingJob


@dataclass
class Session:
    session: boto3.Session
    region: Optional[str] = None

    def _from_arn(self, arn):
        sagemaker = self.session.client("sagemaker", region_name=arn.region)
        logs = self.session.client("logs", region_name=arn.region)

        return Resource.from_arn(arn, sagemaker=sagemaker, cw_logs=logs)

    def _from_name(self, cls, name):
        sagemaker = self.session.client("sagemaker", region_name=self.region)
        logs = self.session.client("logs", region_name=self.region)

        return cls.from_name(name, sagemaker=sagemaker, cw_logs=logs)

    def _from(self, cls, name, arn):
        if arn is not None:
            return self._from_arn(arn)

        if name is not None:
            return self._from_name(cls, name)

        raise ValueError("Either `name` or `arn` need to be passed.")

    def processing_job(self, *, name=None, arn=None):
        self._from(ProcessingJob, name, arn)

    def training_job(self, *, name=None, arn=None):
        self._from(TrainingJob, name, arn)

    def transform_job(self, *, name=None, arn=None):
        self._from(TransformJob, name, arn)

    def hyper_parameter_tuning_job(self, *, name=None, arn=None):
        self._from(HyperParameterTuningJob, name, arn)

    def endpoint(self, *, name=None, arn=None):
        self._from(Endpoint, name, arn)

    def notebook_instance(self, *, name=None, arn=None):
        self._from(NotebookInstance, name, arn)



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Optional

import boto3

# TODO: we could use it from gluon-ts, but this would add a major dependency
from smily.util.settings import Settings, Dependency
from smily.util.aws_config import Config


def sagemaker_client(boto_session):
    return boto_session.client("sagemaker")


def logs_client(boto_session):
    return boto_session.client("logs")


class Env(Settings):
    boto_session: boto3.Session
    sagemaker: Dependency = sagemaker_client
    cw_logs: Dependency = logs_client

    aws_config: Config = Config.read()


env = Env()



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# copied from gluonts.settings

import functools
import inspect
from typing import Any

import pydantic


class Dependency:
    def __init__(self, fn, dependencies):
        self.fn = fn
        self.dependencies = dependencies

    def resolve(self, env):
        kwargs = {key: env[key] for key in self.dependencies}
        return self.fn(**kwargs)


class _Config:
    arbitrary_types_allowed = True


class Settings:
    _cls_types: dict = {}
    _cls_deps: dict = {}

    def __init_subclass__(cls):
        cls._cls_types = {}

        for name, ty in cls.__annotations__.items():
            if ty == Dependency:
                cls._cls_deps[name] = getattr(cls, name)
            else:
                default = getattr(cls, name, ...)
                cls._cls_types[name] = ty, default

    def __init__(self, *args, **kwargs):
        # mapping of key to type, see `_declare` for more info on how this
        # works
        self._types = {}
        self._default = {}
        self._dependencies = {}
        self._context_count = 0

        # We essentially implement our own chainmap, managed by a list. New
        # entries appended to the right; thus, the chain acts as a stack. It is
        # ensured that there are always at least two entries in the chain:
        # A default, used to declare default values for any given key and a
        # base to guard from writing to the default through normal access.
        self._chain = [self._default, kwargs]

        # If sublcassed, `_cls_types` can contain declarations which we need to
        # execute.
        for key, (ty, default) in self._cls_types.items():
            self._declare(key, ty, default=default)

        # Same thing for dependencies.
        for name, fn in self._cls_deps.items():
            self._dependency(name, fn)

    def _reduce(self):
        """"""

        assert not self._context_count, "Cannot reduce within with-blocks."
        compact = {}

        for dct in self._chain[1:]:
            compact.update(dct)

        self._chain = [self._default, compact]

    def _already_declared(self, key):
        return key in self._types or key in self._dependencies

    def _declare(self, key, type=Any, *, default=...):
        assert not self._already_declared(
            key
        ), f"Attempt of overwriting already declared value {key}"

        # This is kinda hacky. For each key, we create a new pydantic model,
        # which contains just one definition, effectively, like this:
        #
        # class foo(pydantic.BaseModel):
        #     foo: type
        #
        # When we want to evaluate, we do this:
        #
        #    # given
        #    settings.foo = value
        #
        #    # becomes
        #    settings._types["foo"].parse_obj({"foo": value}).foo

        self._types[key] = pydantic.create_model(
            key, **{key: (type, ...)}, __config__=_Config
        )

        # we use our own default-handling, instead of relying on pydantic
        if default != ...:
            self._set_(self._default, key, default)

    def _dependency(self, name, fn):
        dependencies = list(inspect.signature(fn).parameters)
        for dependency in dependencies:
            assert self._already_declared(
                dependency
            ), f"`{name}` depends on `{dependency}`, which has not been declared yet."

        self._dependencies[name] = Dependency(fn, dependencies)

    def _get(self, key, default=None):
        """Like `dict.get`."""
        try:
            return self[key]
        except KeyError:
            return default

    def __getitem__(self, key):
        # Iterate all dicts, last to first, and return value as soon as one is
        # found.

        if key in self._dependencies:
            return self._dependencies[key].resolve(self)

        for dct in reversed(self._chain):
            try:
                return dct[key]
            except KeyError:
                pass
        raise KeyError(key)

    def __getattribute__(self, key):
        # We check the key, to check whether we want to acces our chainmap
        # or handle it as a normal attribute.
        if key.startswith("_"):
            return super().__getattribute__(key)
        else:
            return self[key]

    def _set_(self, dct, key, value):
        """Helper method to assign item to a given dictionary.
        Uses `_types` to type-check the value, before assigning.
        """

        assert key not in self._dependencies, "Can't override dependency."

        # If we have type-information, we apply the pydantic-model to the value
        model = self._types.get(key)
        if model is not None:
            value = getattr(model.parse_obj({key: value}), key)

        dct[key] = value

    def __setitem__(self, key, value):
        # Always assigns to the most recent dictionary in our chain.
        self._set_(self._chain[-1], key, value)

    def __setattr__(self, key, value):
        # Same check as in `__getattribute__`.
        if key.startswith("_"):
            super().__setattr__(key, value)
        else:
            self[key] = value

    def _push(self, **kwargs):
        """Add new entry to our chain-map.
        Values are type-checked.
        """
        self._chain.append({})
        # Since we want to type-check, we add the entries manually.
        for key, value in kwargs.items():
            self[key] = value

        return self

    def _pop(self):
        assert len(self._chain) > 2, "Can't pop initial setting."
        return self._chain.pop()

    def __repr__(self):
        inner = ", ".join(list(repr(dct) for dct in self._chain))
        return f"<Settings [{inner}]>"

    def _let(self, **kwargs) -> "_ScopedSettings":
        """Create a new context, where kwargs are added to the chain::
            with settings._let(foo=42):
                assert settings.foo = 42
        `_let` does not push a new context, but returns a `_ScopedSettings`
        object, that pushes the context, when entered through a
        `with`-statement.
        """
        return _ScopedSettings(self, kwargs)

    def _inject(self, *keys):
        """Dependency injection.
        This will inject values from settings if avaiable and not passed
        directly::
            @settings._inject("foo")
            def fn(foo=1):
                return foo
            # Since foo is not available in settings, the functions default
            # value is taken.
            assert fn() == 1
            with settings._let(foo=2):
                # Since foo is declared in the settings, it is used.
                assert fn() == 2
                # Directly passed values always take precedence.
                assert fn(3) == 3
        """

        def dec(fn):
            # We need the signature to be able to assemble the args later.
            sig = inspect.signature(fn)

            for key in keys:
                assert key in sig.parameters, f"Key {key} not in arguments."

            @functools.wraps(fn)
            def wrapper(*args, **kwargs):
                # arguments are always keyword params
                arguments = sig.bind_partial(*args, **kwargs).arguments

                setting_kwargs = {
                    key: self[key] for key in keys if key not in arguments
                }

                return fn(**arguments, **setting_kwargs)

            return wrapper

        return dec


class _ScopedSettings:
    def __init__(self, settings, kwargs):
        self.settings = settings
        self.kwargs = kwargs

    def __enter__(self):
        self.settings._context_count += 1
        return self.settings._push(**self.kwargs)

    def __exit__(self, *args):
        self.settings._context_count -= 1
        self.settings._pop()


def let(settings, **kwargs):
    "`let(settings, ...)` is the same as `settings._let(...)`."
    return settings._let(**kwargs)


def inject(settings, *args, **kwargs):
    "`inject(settings, ...)` is the same as `settings._inject(...)`."
    return settings._inject(*args, **kwargs)



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import NamedTuple, Optional

from toolz import first, second


class Arn(NamedTuple):
    # arn:partition:service:region:account-id:<resource>`
    # where resource is one of
    # [resource-id, resource-type/resource-id, resource-type:resource-id]
    service: str
    region: str
    account: str
    resource: str

    @staticmethod
    def parse(arn: str) -> "Arn":
        # we can ignore the leading `arn:aws`
        elements = arn.split(":", 5)[2:]
        kwargs = dict(
            zip(["service", "region", "account", "resource"], elements)
        )

        return Arn(**kwargs)

    @classmethod
    def __get_validators__(cls):
        # For pydantic
        yield cls.parse

    def resource_split(self):
        """Return a pair of (resource_type, resource_id).

        For example, an s3-path would return:

            <bucket>, <key>

        However, in cases where the resource can't be split, it returns the
        resource twice.

        Given just a s3-bucket-arn, we get:

            <bucket>, <bucket>
        """
        for sep in ":", "/":
            if sep in self.resource:
                return self.resource.split(sep, 1)

        return self.resource, self.resource

    @property
    def resource_type(self):
        return first(self.resource_split())

    @property
    def resource_id(self):
        return second(self.resource_split())

    def __str__(self):
        return (
            "arn:aws:"
            f"{self.service}:{self.region}:{self.account}:{self.resource}"
        )



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import re
from typing import List

from toolz import pluck, dissoc

from smily import env
from smily.util import ignore_nones

from .arn import Arn
from .logs import LogStream, LogStreamFilter, TextFilter


def snake_case(s, sep="_"):
    """Tranform a string from CamelCase to snake_case.

    >>> snake_case("TrainingJob")
    "training_job"

    """
    return re.sub(r"(?<!^)(?=[A-Z])", sep, s).lower()


class Resource:
    _resource_map = {}

    def __init_subclass__(cls, *args, **kwargs):
        cls._resource_map[cls.arn_resource_name()] = cls
        return cls

    @classmethod
    def class_for_arn(cls, arn):
        return cls._resource_map[arn.resource_type]

    @classmethod
    @env._inject("sagemaker", "cw_logs")
    def from_arn(cls, arn, *, sagemaker, cw_logs):
        if isinstance(arn, str):
            arn = Arn.parse(arn)

        ty = cls.class_for_arn(arn)

        return ty.from_name(arn.resource_id, sagemaker=sagemaker, cw_logs=cw_logs)

    @classmethod
    def kind(cls):
        # E.g. `TrainingJob` for `class TrainingJob`
        return cls.__name__

    @classmethod
    def arn_resource_name(cls):
        # E.g. `training-job` for `class TrainingJob`
        return snake_case(cls.kind(), "-")

    @classmethod
    def arn_field(cls):
        # E.g. `TrainingJobArn` for `class TrainingJob`
        return f"{cls.kind()}Arn"

    @classmethod
    def name_field(cls):
        # E.g. `TrainingJobName` for `class TrainingJob`
        return f"{cls.kind()}Name"

    @classmethod
    def Arn(cls, name, *, account, region):
        return Arn(
            service="sagemaker",
            region=region,
            account=account,
            resource=f"{cls.arn_resource_name()}/{name}",
        )

    @classmethod
    def _describe(cls, name, sagemaker):
        describe_op = f"describe_{snake_case(cls.kind())}"
        describe_fn = getattr(sagemaker, describe_op)
        return describe_fn(**{cls.name_field(): name})

    def log_group_name(self):
        return f"/aws/sagemaker/{self.kind()}s"

    @classmethod
    @env._inject("sagemaker", "cw_logs")
    def from_name(cls, name, *, sagemaker, cw_logs):
        "Calls `describe` using `name`."
        return cls(
            cls._describe(name, sagemaker),
            sagemaker=sagemaker,
            cw_logs=cw_logs,
        )

    def __init__(self, description, *, sagemaker: object, cw_logs: object):
        self._description = description
        self._sagemaker = sagemaker
        self._cw_logs = cw_logs
        self._log_streams = None

    @property
    def arn(self):
        return self._description[self.arn_field()]

    @property
    def name(self):
        return self._description[self.name_field()]

    def description(self, full=False):
        omit_fields = ["ResponseMetadata"]
        if not full:
            omit_fields += self.description_omit_fields()

        description = dissoc(self._description, *omit_fields)
        return self.postprocess_description(description)

    def description_omit_fields(self):
        return []

    def postprocess_description(self, description):
        return description

    def log_stream_name_prefix(self):
        return self.name + "/"

    def _describe_log_streams(self, nextToken=None):
        return ignore_nones(self._cw_logs.describe_log_streams)(
            logGroupName=self.log_group_name(),
            logStreamNamePrefix=self.log_stream_name_prefix(),
            nextToken=nextToken,
        )

    def _get_log_streams(self):
        if self._log_streams is None:
            self._log_streams = {}
            token = None
            while True:
                response = self._describe_log_streams(token)
                token = response.get("nextToken")
                for stream in response["logStreams"]:
                    self._log_streams[stream["logStreamName"]] = stream
                if token is None:
                    break
        return self._log_streams

    def _log_stream_names(self):
        streams = self._get_log_streams()
        return list(streams)

    @property
    def logs(self) -> List[LogStream]:
        return [
            LogStream(self._cw_logs, self.log_group_name(), stream_name)
            for stream_name in self._log_stream_names()
        ]

    @property
    def logfilter(self) -> LogStreamFilter:
        return LogStreamFilter(
            logs=self._cw_logs,
            group=self.log_group_name(),
            stream_names=self._log_stream_names(),
        )


class TextResource:
    def __init__(self, text):
        self.text = text

    @property
    def logfilter(self):
        return TextFilter(self.text)


class Endpoint(Resource):
    url_prefix = "endpoints"

    def log_stream_name_prefix(self):
        """In contrast to every other Resource every Endpoint gets its own
        LogGroup. Thus, all LogStreams are returned for an Endpoint.
        """
        return None

    def log_group_name(self):
        return f"/aws/sagemaker/Endpoints/{self.name}"

    def description_omit_fields(self):
        return ["CreationTime", "LastModifiedTime"]


class NotebookInstance(Resource):
    url_prefix = "notebook-instances"


class TrainingJob(Resource):
    url_prefix = "jobs"

    def description_omit_fields(self):
        return [
            "SecondaryStatusTransitions",
            "StoppingCondition",
            "RoleArn",
            "CreationTime",
            "TrainingStartTime",
            "TrainingEndTime",
            "LastModifiedTime",
            "EnableNetworkIsolation",
            "EnableInterContainerTrafficEncryption",
            "EnableManagedSpotTraining",
            "BillableTimeInSeconds",
            "OutputDataConfig",  # same info as ModelArtifacts
        ]

    def postprocess_description(self, description):
        description["InputDataConfig"] = {
            channel["ChannelName"]: channel["DataSource"]["S3DataSource"]["S3Uri"]
            for channel in description["InputDataConfig"]
        }

        if "ModelArtifacts" in description:
            description["ModelArtifacts"] = description["ModelArtifacts"][
                "S3ModelArtifacts"
            ]
        description["TrainingImage"] = description.pop("AlgorithmSpecification")[
            "TrainingImage"
        ]

        if "FinalMetricDataList" in description:
            description["FinalMetricDataList"] = {
                metric["MetricName"]: metric["Value"]
                for metric in description["FinalMetricDataList"]
            }

        description["ResourceConfig"] = dissoc(
            description["ResourceConfig"], "VolumeSizeInGB", "VolumeKmsKeyId"
        )
        return dict(sorted(description.items()))


class ProcessingJob(Resource):
    url_prefix = "processing-jobs"


class TransformJob(Resource):
    url_prefix = "transform-jobs"


class HyperParameterTuningJob(Resource):
    url_prefix = "hyper-tuning-jobs"

    def log_stream_name_prefix(self):
        return self._description[self.name_field()]

    def log_group_name(self):
        return "/aws/sagemaker/TrainingJobs"



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import re

import boto3
import click

import smily


class ArnParamType(click.ParamType):
    name = "arn"

    def convert(self, value, param, ctx):
        try:
            return smily.Arn.parse(value)

        except AssertionError:
            self.fail(f"{value!r} is not a valid arn", param, ctx)


ArnType = ArnParamType()


class ResourceParamType(click.ParamType):
    name = "resource"

    def convert(self, value, param, ctx):
        try:
            arn = smily.Arn.parse(value)

            profile = smily.env.aws_config.accounts.get(arn.account)

            smily.env._push(
                boto_session=boto3.Session(profile_name=profile, region_name=arn.region)
            )

            return smily.Resource.from_arn(arn)

        except AssertionError:
            self.fail(f"{value!r} is not a valid arn", param, ctx)


Resource = ResourceParamType()


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])

        ctx.fail(
            f"Too many matches `{cmd_name}` could mean: \n"
            + "\n".join([f"  * {match}" for match in sorted(matches)])
        )


@click.group(cls=AliasedGroup)
def main():
    pass



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import functools

from toolz import valfilter

from .settings import Settings

__all__ = ["ignore_nones", "Settings"]


def ignore_nones(fn):
    """Decorator which calls `fn` with `*args, **kwargs`, but pops all items of
    `kwargs`, which are `None` (`foo(bar=None)` becomes `foo()`).

    This is useful when using boto3, since it often doesn't allow passing
    `None` as a not-specified value.

    ::

        @ignore_nones
        def foo(bar=42):
            return bar

        assert foo(None) == 42
    """

    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        kwargs = valfilter(lambda val: val is not None, kwargs)
        return fn(*args, **kwargs)

    return wrapper



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

__all__ = [
    "env",
    "Arn",
    "Resource",
    "Endpoint",
    "NotebookInstance",
    "TrainingJob",
    "ProcessingJob",
    "TransformJob",
    "HyperParameterTuningJob",
    "arn",
    "Session",
]

from ._env import env

from .lib import (
    Arn,
    Resource,
    Endpoint,
    NotebookInstance,
    TrainingJob,
    ProcessingJob,
    TransformJob,
    HyperParameterTuningJob,
)


from .session import Session
from ._smile import arn



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import webbrowser
from urllib.parse import quote_plus

import click

from smily.lib import Resource
from ._base import main, ArnType


@main.command()
@click.argument("arn", required=True, type=ArnType)
@click.option("--print", "-p", "print_", is_flag=True)
def open(arn, print_):
    """
    Open a SageMaker training-job or HPO-job in the browser:

        qi open arn:aws:sagemaker:us-east-1:123456789012:training-job/job-name

    The -p/--print option just prints the link, instead of opening it.
    """

    ResourceKind = Resource.class_for_arn(arn)

    destination = (
        f"sagemaker/home?region={arn.region}"
        f"#/{ResourceKind.url_prefix}/{arn.resource_id}"
    )

    url = f"https://{arn.region}.console.aws.amazon.com/{destination}"

    if print_:
        click.echo(url)
    else:
        webbrowser.open(url)



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from setuptools import setup, find_packages

setup(
    name="smily",
    author="Amazon",
    version="0.0.4",
    author_email="gluon-ts-dev@amazon.com",
    maintainer_email="gluon-ts-dev@amazon.com",
    packages=find_packages("."),
    install_requires=[
        "boto3",
        "toolz",
        "pydantic",
    ],
    extras_require={"cli": ["click", "rich"]},
    entry_points={
        "console_scripts": ["smily=smily.__main__:main"],
    },
)



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import click

from ._base import main, AliasedGroup, Resource


@main.group(cls=AliasedGroup)
def logs():
    """Interact with cloudwatch logs of a training-job, processing-job,
    endpoint, hyperparameter-tuning-job, notebook-instance or a
    transform-job."""
    pass


@logs.command()
@click.argument("arn", required=True, type=Resource)
@click.option("-n", type=int)
def head(arn, n):
    for log in arn.logs:
        click.secho(log.stream_name, bold=True, color="blue", err=True)
        for line in log.head(n=n):
            click.echo(line)
        click.echo()


@logs.command()
@click.argument("arn", required=True, type=Resource)
@click.option("-n", type=int, default=10)
@click.option("-f", "follow", is_flag=True)
@click.option("-i", "--interval", default=10)
def tail(arn, n, follow, interval):
    for log in arn.logs:
        click.secho(log.stream_name, bold=True, color="blue", err=True)
        if follow:
            lines = log.follow(limit=n, start_at_top=False, interval=interval)
        else:
            lines = log.tail(n)

        for line in lines:
            click.echo(line)
        click.echo()


@logs.command()
@click.argument("arn", required=True, type=Resource)
@click.option("-f", "follow", is_flag=True)
@click.option("-i", "--interval", default=10)
def cat(arn, follow, interval):
    for log in arn.logs:
        click.secho(log.stream_name, bold=True, color="blue", err=True)

        if follow:
            lines = log.follow(interval=interval)
        else:
            lines = log.cat()

        for line in lines:
            click.echo(line)
        click.echo()


@logs.command()
@click.option("--expression", "-e", required=True)
@click.argument("arn", required=True, type=Resource)
def filter(arn, expression):
    for name, logfilterevents in arn.logfilter.filter(expression).items():
        click.secho(name, bold=True, color="blue", err=True)
        for evt in logfilterevents:
            click.secho(evt.message)
        click.echo()



import pandas as pd
from pytimeparse.timeparse import timeparse

from . import operator as op


class Base:
    def convert(self, value):
        return value

    def __iter__(self):
        yield op.has(self.name)

    def __invert__(self):
        return [op.hasnot(self.name)]


class Equals(Base):
    def __eq__(self, other):
        return [op.eq(self.name, self.convert(other))]


class NotEquals(Base):
    def __ne__(self, other):
        return [op.ne(self.name, self.convert(other))]


class Equality(Equals, NotEquals):
    pass


class GreaterThan(Base):
    def __gt__(self, other):
        return [op.gt(self.name, self.convert(other))]


class GreaterEquals(Base):
    def __ge__(self, other):
        return [op.ge(self.name, self.convert(other))]


class LessThan(Base):
    def __lt__(self, other):
        return [op.lt(self.name, self.convert(other))]


class LessEquals(Base):
    def __le__(self, other):
        return [op.le(self.name, self.convert(other))]


class Contains(Base):
    def contains(self, other):
        return [op.contains(self.name, self.convert(other))]


class In(Base):
    def one_of(self, args):
        return [op.in_(self.name, ",".join(args))]


class Ordered(Equals, NotEquals, GreaterThan, GreaterEquals, LessThan, LessEquals):
    pass


class Any(Ordered, Contains, In):
    pass


class DictField(Ordered):
    def __init__(self, name):
        self.name = name

    def __iter__(self):
        yield op.has(self.name)

    def __invert__(self):
        return [op.hasnot(self.name)]


class Dict:
    def __getattr__(self, name):
        return DictField(f"{self.name}.{name}")


class String(Equality, Contains, In):
    pass


class Number(Ordered):
    pass


class Duration(Number):
    def convert(self, value):
        val = timeparse(value)
        if val is not None:
            return str(val)

        return value


class Timestamp(Number):
    def convert(self, value):
        t = pd.Timestamp(value)
        if t.tzinfo:
            t = t.tz_convert("UTC")
        return t.strftime("%Y-%m-%dT%H:%M:%SZ")
